import ctypes
import random
from   sdl2   import *
import sys
import time

def usage():
    print('''The Pong Game.
Keys:
  W:         move left paddle up
  S:         move left paddle down
  Up:        move right paddle up
  Down:      move right paddle down
  1:         toggle AI for left paddle
  0:         toggle AI for right paddle
  Backspace: reset scores
''')

FPS = 60

class Vec2:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __floordiv__(self, x):
        return Vec2(self.x // x, self.y // x)

    def __iadd__(self, v):
        self.x += v.x
        self.y += v.y
        return self

    def __isub__(self, v):
        self.x -= v.x
        self.y -= v.y
        return self

resolution = Vec2(800, 600)

class Ticking:
    objects = []
    def __init__(self):
        super().__init__()
        Ticking.objects.append(self)

    def tick(self):
        pass

    def destroy(self):
        Ticking.objects.remove(self)

    @staticmethod
    def tick_all():
        for object in Ticking.objects:
            object.tick()

class Renderable:
    objects = []
    def __init__(self):
        super().__init__()
        Renderable.objects.append(self)

    def render(self):
        pass

    def destroy(self):
        super().destroy()
        Renderable.objects.remove(self)

    @staticmethod
    def render_all(renderer):
        for object in Renderable.objects:
            object.render(renderer)

class Entity(Ticking, Renderable):
    def __init__(self, position, size):
        super().__init__()
        self.halfsize = size // 2
        self.position = position
        self.velocity = Vec2(0, 0)
        self.rect = SDL_Rect(w = size.x, h = size.y)

    def render(self, renderer):
        SDL_RenderFillRect(renderer, self.rect)

    def tick(self):
        self.position += self.velocity
        self.rect.x = self.position.x - self.halfsize.x
        self.rect.y = self.position.y - self.halfsize.y

class Paddle(Entity):
    def __init__(self, left):
        width  = 20
        height = 100
        if left:
            x = width // 2
        else:
            x = resolution.x - width // 2
        super().__init__(Vec2(x, resolution.y // 2), Vec2(width, height))
        self.score = 0

    def tick(self):
        super().tick()
        self.position.y = clamp(
                self.position.y, self.halfsize.y, resolution.y - self.halfsize.y
        )

    def move_up(self):
        self.velocity.y = -3

    def move_down(self):
        self.velocity.y = 3

    def stop(self):
        self.velocity.y = 0

class Ball(Entity):
    def __init__(self):
        super().__init__(resolution // 2, Vec2(20, 20))

    def reset(self):
        self.position = resolution // 2
        if random.randint(0, 1) == 0:
            self.velocity.x = -3
        else:
            self.velocity.x = 3
        self.velocity.y = random.randint(-5, 5)

class AI(Ticking):
    def __init__(self, paddle, ball):
        super().__init__()
        self.paddle = paddle
        self.ball   = ball

    def tick(self):
        def move_to(y):
            if self.paddle.position.y < y:
                self.paddle.move_down()
            elif self.paddle.position.y > y:
                self.paddle.move_up()
            else:
                self.paddle.stop()
        if (self.ball.velocity.x > 0) == \
                (self.ball.position.x > self.paddle.position.x):
            move_to(resolution.y // 2)
        else:
            move_to(self.ball.position.y)

class Board(Renderable):
    def __init__(self, renderer):
        super().__init__()
        digits = '''
***   * *** *** * * *** *** *** *** ***
* *   *   *   * * * *   *     * * * * *
* *   * *** *** *** *** ***   * *** ***
* *   * *     *   *   * * *   * * *   *
***   * *** ***   * *** ***   * *** ***
'''
        self.scale = 5

        self.texture = SDL_CreateTexture(
                renderer,
                SDL_GetWindowPixelFormat(window),
                SDL_TEXTUREACCESS_TARGET,
                len(digits) // 5 * self.scale,
                5 * self.scale
        )
        SDL_SetRenderTarget(renderer, self.texture)
        SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255)
        SDL_RenderClear(renderer)

        SDL_SetRenderDrawColor(renderer, 128, 128, 128, 255)
        rect = SDL_Rect(w = self.scale, h = self.scale)
        x = 0
        y = -self.scale
        for c in digits:
            if c == '*':
                rect.x = x
                rect.y = y
                SDL_RenderFillRect(renderer, rect)
                x += self.scale
            elif c == '\n':
                y += self.scale
                x = 0
            else:
                x += self.scale

        SDL_SetRenderTarget(renderer, None)

        self.source      = SDL_Rect(y = 0, w = 3 * self.scale, h = 5 * self.scale)
        self.destination = SDL_Rect(y = 0, w = self.source.w,  h = self.source.h)

        self.reset()

    def score(self, player):
        self._score[player] += 1
        self._text[player] = [ ord(d) - ord('0') for d in str(self._score[player]) ]

    def reset(self):
        self._score = [ 0, 0 ]
        self._text  = [ [ 0 ], [ 0 ] ]

    def render(self, renderer):
        def render_digit(digit, x):
            self.source.x = digit * 4 * self.scale
            self.destination.x = x
            SDL_RenderCopy(renderer, self.texture, self.source, self.destination)
        def render_score(player, x):
            for digit in self._text[player]:
                render_digit(digit, x)
                x += 4 * self.scale
        render_score(
                0, resolution.x // 2 - 30 - len(self._text[0]) * 4 * self.scale
        )
        render_score(1, resolution.x // 2 + 30)

def clamp(value, min, max):
    if value < min:
        return min
    if value > max:
        return max
    return value

def intersect(entity1, entity2):
    return SDL_HasIntersection(entity1.rect, entity2.rect)

def hit(paddle, ball):
    degrees = 0.7
    step = paddle.halfsize.y // 10
    factor = (ball.position.y - paddle.position.y) // step
    ball.velocity.x = -ball.velocity.x
    ball.velocity.y = int(factor * degrees) + paddle.velocity.y

if __name__ == '__main__':
    if len(sys.argv) > 1 and (sys.argv[1] == '-h' or sys.argv[1] == '--help'):
        usage()
        sys.exit(0)


    SDL_Init(SDL_INIT_VIDEO)
    window = SDL_CreateWindow(
            b'Pong',
            SDL_WINDOWPOS_UNDEFINED,
            SDL_WINDOWPOS_UNDEFINED,
            resolution.x,
            resolution.y,
            SDL_WINDOW_SHOWN
    )

    renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED)

    board = Board(renderer)
    paddle1 = Paddle(True)
    paddle2 = Paddle(False)
    ball    = Ball()
    ball.reset()

    ai = [ None, None ]
    def toggle_ai(index, paddle):
        if ai[index]:
            ai[index].destroy()
            ai[index] = None
            paddle.velocity.y = 0
        else:
            ai[index] = AI(paddle, ball)

    frame_time = 1 / FPS
    running = True
    event = SDL_Event()
    rect = SDL_Rect()
    while running:
        frame_start = time.monotonic()

        while SDL_PollEvent(ctypes.byref(event)) != 0:
            if event.type == SDL_QUIT:
                running = False
                break
            if event.type == SDL_KEYDOWN:
                key = event.key.keysym.sym
                if key == SDLK_w:
                    if not ai[1]: paddle1.move_up()
                elif key == SDLK_s:
                    if not ai[1]: paddle1.move_down()
                elif key == SDLK_UP:
                    if not ai[0]: paddle2.move_up()
                elif key == SDLK_DOWN:
                    if not ai[0]: paddle2.move_down()
                elif key == SDLK_0:
                    toggle_ai(0, paddle2)
                elif key == SDLK_1:
                    toggle_ai(1, paddle1)
                elif key == SDLK_BACKSPACE:
                    board.reset()
            elif event.type == SDL_KEYUP:
                key = event.key.keysym.sym
                if key in (SDLK_w, SDLK_s):
                    paddle1.stop()
                elif key in (SDLK_UP, SDLK_DOWN):
                    paddle2.stop()

        Ticking.tick_all()

        if intersect(ball, paddle1):
            hit(paddle1, ball)
        elif intersect(ball, paddle2):
            hit(paddle2, ball)

        if ball.position.x - ball.halfsize.x < 0:
            board.score(1)
            ball.reset()
        elif ball.position.x + ball.halfsize.x > resolution.x:
            board.score(0)
            ball.reset()
        elif ball.position.y - ball.halfsize.y < 0:
            ball.velocity.y = -ball.velocity.y
            ball.position.y = 2 * ball.halfsize.y - ball.position.y
        elif ball.position.y + ball.halfsize.y > resolution.y:
            ball.velocity.y = -ball.velocity.y
            ball.position.y = 2 * (resolution.y - ball.halfsize.y) - ball.position.y

        SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255)
        SDL_RenderClear(renderer)

        SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255)

        Renderable.render_all(renderer)

        SDL_RenderPresent(renderer)

        frame_end = time.monotonic()
        delay = frame_time - (frame_end - frame_start)
        if delay > 0: SDL_Delay(int(delay * 1000))

    SDL_DestroyWindow(window)
    SDL_Quit()
